import httpx
import requests
from httpx import Headers, Request, Response, Auth
from httpx._client import Client, logger
from requests.auth import AuthBase


class HttpxToRequestsAuth(AuthBase):
    def __init__(self, httpx_auth):
        self.httpx_auth: Auth = httpx_auth

    def __call__(self, request: Request):

        # Create a mock HTTPX request
        httpx_request = httpx.Request(
            method=request.method,
            url=request.url,
            headers=request.headers
        )

        # Apply the httpx authentication
        if self.httpx_auth:
            self.httpx_auth.auth_flow(httpx_request)

        # Copy back modified headers to the requests request
        request.headers.update(httpx_request.headers)

        return request


def _send_single_request(self: Client, request: Request) -> Response:
    auth_patched = HttpxToRequestsAuth(self.auth)

    # Send the request using requests
    resp = requests.request(
        method=request.method,
        url=request.url,
        headers=request.headers,
        data=request.content,
        auth=auth_patched,
        stream=True
    )

    # If the response is not successful, raise an HTTPError
    if not resp.ok:
        raise httpx.HTTPStatusError(
            f"HTTP request failed with status code {resp.status_code}",
            request=request,
            response=resp
        )

    # Read the content of the response and decode it
    content = resp.raw.read(decode_content=True)

    # Remove content-encoding header to prevent httpx from trying to decode the content
    # since requests already decodes the content
    if "content-encoding" in resp.headers:
        del resp.headers["content-encoding"]

    # Transform the requests response to the httpx response
    response = Response(
        status_code=resp.status_code,
        headers=Headers(resp.headers),
        content=content,
        request=httpx.Request(method=resp.request.method, url=str(resp.url)),
    )

    logger.debug(
        'HTTP Request: %s %s "%s %d %s"',
        request.method,
        request.url,
        response.http_version,
        response.status_code,
        response.reason_phrase,
    )

    return response


def pyodide_patch_httpx():
    """
    Patch the httpx.Client to use requests to send requests
    """
    Client._send_single_request = _send_single_request



import json
import logging
import os
import warnings
from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, Union, Optional

import boto3

if TYPE_CHECKING:
    from pandas import DataFrame

from cirro.models.s3_path import S3Path

logger = logging.getLogger(__name__)


def _fix_s3_path(path: str) -> str:
    """
    Fix the S3 path to ensure it starts with 's3://'.
    """
    normalized_path = path.replace(os.sep, '/').strip()
    if normalized_path.startswith("s3:/") and not normalized_path.startswith("s3://"):
        return normalized_path.replace("s3:/", "s3://", 1)
    return path


def write_json(dat, local_path: str, indent=4):
    """Write a JSON object to a local file."""
    with Path(local_path).open(mode="wt") as handle:
        return json.dump(dat, handle, indent=indent)


def read_csv(path: str, required_columns=None) -> 'DataFrame':
    """Read a CSV from the dataset and check for any required columns."""
    if required_columns is None:
        required_columns = []
    path = _fix_s3_path(path)
    import pandas as pd
    df = pd.read_csv(path)
    for col in required_columns:
        assert col in df.columns.values, f"Did not find expected columns {col} in {path}"
    return df


def read_json(path: str):
    """Read a JSON object from a local file or S3 path."""
    path = _fix_s3_path(path)
    s3_path = S3Path(path)

    if s3_path.valid:
        s3 = boto3.client('s3')
        retr = s3.get_object(Bucket=s3_path.bucket, Key=s3_path.key)
        text = retr['Body'].read().decode()
    else:
        with Path(path).open() as handle:
            text = handle.read()

    # Parse JSON
    return json.loads(text)


class PreprocessDataset:
    """
    Helper functions for performing preparatory tasks immediately before launching
    the analysis workflow for a dataset.

    For use in the `preprocess.py` script.
    More info: https://docs.cirro.bio/pipelines/preprocess-script/
    """
    samplesheet: 'DataFrame'
    """
    A pandas DataFrame containing all of the metadata assigned to the samples present
    in the input datasets (at the time of analysis).

    More info: https://docs.cirro.bio/pipelines/preprocess-script/#dssamplesheet
    """
    files: 'DataFrame'
    """
    A DataFrame containing information on the files contained in the input datasets,
    and the sample that each file is assigned to.

    More info: https://docs.cirro.bio/pipelines/preprocess-script/#dsfiles
    """
    params: dict
    """
    A dictionary with all of the parameter values populated by user input
    using the process-form.json and process-input.json configurations.

    This is read-only, use `add_param` to add new parameters or `remove_param` to remove them.

    More info: https://docs.cirro.bio/pipelines/preprocess-script/#dsparams
    """
    metadata: dict
    """
    Detailed information about the dataset at the time of analysis,
    including the project, process, and input datasets.

    More info: https://docs.cirro.bio/pipelines/preprocess-script/#dsmetadata
    """
    dataset_root: str
    """
    Base path to the dataset
    """

    _PARAMS_FILE = "params.json"
    _REFERENCES_BASE = "s3://pubweb-references"

    def __init__(self,
                 samplesheet: Union['DataFrame', str, Path],
                 files: Union['DataFrame', str, Path],
                 params: dict = None,
                 metadata: dict = None,
                 dataset_root: str = None):
        import pandas as pd
        # Convert DataFrame to string if necessary
        if isinstance(samplesheet, str):
            samplesheet = pd.read_csv(StringIO(samplesheet))
        if isinstance(samplesheet, Path):
            samplesheet = read_csv(str(samplesheet))
        if isinstance(files, str):
            files = pd.read_csv(StringIO(files))
        if isinstance(files, Path):
            files = read_csv(str(files))
        if params is None:
            params = {}
        if metadata is None:
            metadata = {}

        self.samplesheet = samplesheet
        self.files = files
        self.params = params
        self.metadata = metadata
        self.dataset_root = dataset_root

    @classmethod
    def from_path(cls, dataset_root: str, config_directory='config'):
        """
        Creates an instance from a path
        (useful for testing or when running the script outside Cirro)
        """
        config_directory = Path(dataset_root, config_directory)

        files = read_csv(
            str(Path(config_directory, "files.csv")),
            required_columns=["sample", "file"]
        )

        samplesheet = read_csv(
            str(Path(config_directory, "samplesheet.csv")),
            required_columns=["sample"]
        )

        params = read_json(
            str(Path(config_directory, "params.json")),
        )

        metadata = read_json(
            str(Path(config_directory, "metadata.json")),
        )

        return cls(files=files,
                   samplesheet=samplesheet,
                   params=params,
                   metadata=metadata,
                   dataset_root=dataset_root)

    @classmethod
    def from_running(cls):
        """
        Creates an instance from the currently running dataset
        (expected to be called from inside a Cirro analysis process)
        """
        logging.basicConfig(level=logging.INFO,
                            format='%(asctime)s %(levelname)-8s [PreprocessDataset] %(message)s')
        dataset_path = os.getenv("PW_S3_DATASET")
        return cls.from_path(dataset_path)

    @property
    def references_base(self):
        """
        Returns the base URL for references.
        This is used to access public references in the Cirro system.
        """
        return self._REFERENCES_BASE

    def log(self):
        """Print logging messages about the dataset."""
        logger.info(f"Storage location for dataset: {self.dataset_root}")
        logger.info(f"Number of files in dataset: {self.files.shape[0]:,}")
        logger.info(f"Number of samples in dataset: {self.samplesheet.shape[0]:,}")

    def add_param(self, name: str, value, overwrite=False, log=True):
        """Add a parameter to the dataset."""

        assert overwrite or name not in self.params, \
            f"Cannot add parameter {name}, already exists (and overwrite=False)"

        if log:
            logger.info(f"Adding parameter {name} = {value}")
        self.params[name] = value

        if log:
            logger.info("Saving parameters")
        write_json(self.params, self._PARAMS_FILE)

    def remove_param(self, name: str, force=False):
        """Remove a parameter from the dataset."""

        assert force or name in self.params, \
            f"Cannot remove parameter {name}, does not exist (and force=False)"

        logger.info(f"Removing parameter {name}")
        if name in self.params:
            del self.params[name]

        logger.info("Saving parameters")
        write_json(self.params, self._PARAMS_FILE)

    def keep_params(self, params_to_keep: list[str]):
        """Keep only the specified parameters in the dataset."""
        logger.info(f"Keeping parameters: {params_to_keep}")
        self.params = {
            k: v for k, v in self.params.items()
            if k in params_to_keep
        }
        write_json(self.params, self._PARAMS_FILE)

    def update_compute(self, from_str: str, to_str: str, fp="nextflow-override.config"):
        """Replace all instances of a text string in the compute config file."""

        assert os.path.exists(fp), f"File does not exist: {fp}"
        with open(fp, 'r') as handle:
            compute = handle.read()
        n = len(compute.split(from_str)) - 1
        logger.info(f"Replacing {n:,} instances of {from_str} with {to_str} in {fp}")
        compute = compute.replace(from_str, to_str)
        with open(fp, 'wt') as handle:
            handle.write(compute)

    def pivot_samplesheet(
            self,
            index=None,
            pivot_columns: Union[Optional[str], list[str]] = 'read',
            metadata_columns: list[str] = None,
            column_prefix: str = "fastq_",
            file_filter_predicate: str = None
    ):
        """
        Combines data from both the samples and files table into a wide format with
        each sample on a row and each file in a column.
        The file column indexes are created by default from the `read` column, but can be customized.

        For example, if the `files` table has columns `sample`, `read`, and `file`,
        and the `samplesheet` has columns `sample`, `status`, and `group`, the output
        will have columns `sample`, `fastq_1`, `fastq_2`, `status`, and `group`.

        Args:
            index: List[str], used to make the frames new index, defaults to
             `["sampleIndex", "sample", "lane"]`
            pivot_columns: str or List[str], columns to pivot on and create the new column,
             defaults to 'read'. This effectively makes the column `<column_prefix><read>'.
             If the column is not defined or not present, the pivot column will be generated
             from the file number index.
            metadata_columns: List[str], metadata columns to include in the output,
             defaults to all columns that are available from the sample metadata.
             If your pipeline doesn't like extra columns, make sure to specify the allowed columns here.
            column_prefix: str, optional, prefix for the new columns, defaults to `fastq_`.
            file_filter_predicate: str, optional, a pandas query string to filter the files table.
             A common use case would be to filter out indexed reads, e.g. `readType == "R"`.

        Returns:
            DataFrame: A wide-format sample sheet with the specified columns pivoted.
        """
        import pandas as pd

        pivoted_files = self.pivot_files(index=index,
                                         pivot_columns=pivot_columns,
                                         column_prefix=column_prefix,
                                         file_filter_predicate=file_filter_predicate)
        combined = pd.merge(pivoted_files, self.samplesheet, on='sample', how="inner", validate="many_to_many")

        # Default to keeping all columns
        if metadata_columns is None:
            metadata_columns = self.samplesheet.columns.tolist() + pivoted_files.columns.tolist()

        # Keep only the specified metadata columns
        all_columns = combined.axes[1]
        for column in all_columns:
            if (column not in metadata_columns
                    # These columns are required, never drop them
                    and column_prefix not in column
                    and 'sample' != column):
                combined = combined.drop(columns=[column])

        return combined

    def pivot_files(
            self,
            index: list[str] = None,
            pivot_columns: Union[str, list[str]] = 'read',
            column_prefix: str = "fastq_",
            file_filter_predicate: str = None
    ):
        """
        Format the files table into a wide format with each sample on a row
        and each file in a column. The column indexes are created by default
        from the `read` column, but can be customized. This is useful for
        paired-end sequencing data where you want to have the columns
        `sample`, `fastq_1`, and `fastq_2` as the output.

        Args:
            index: List[str], used to make the frames new index, defaults to
            pivot_columns: str or List[str], columns to pivot on and create the new column,
             defaults to 'read'. This effectively makes the column `<column_prefix><read>`
            column_prefix: str, optional, prefix for the new columns, defaults to `fastq_`.
            file_filter_predicate: str, optional, a pandas query string to filter the files table.

        Returns:
            DataFrame: A wide-format sample sheet with the specified columns pivoted.
        """
        if index is None:
            index = ["sampleIndex", "sample", "lane"]
        logger.info("Formatting a wide files table")
        logger.info("File table (long)")
        logger.info(self.files.head().to_csv(index=False))

        files = self.files

        if file_filter_predicate is not None:
            # Filter the files table based on the predicate
            files = files.query(file_filter_predicate)

        # If we don't have access to the column defined, just use the file number
        # By default this is 'read' but the data might not be paired
        pivot_columns_defined = pivot_columns is not None and len(pivot_columns) > 0
        if not pivot_columns_defined or pivot_columns not in files.columns.values:
            logger.warning("Pivot column not found, grouping by sample instead.")
            files['file_num'] = files.groupby('sample').cumcount() + 1
            pivot_columns = 'file_num'

        if isinstance(pivot_columns, str):
            pivot_columns = [pivot_columns]

        assert pivot_columns in files.columns.values, f"Column '{pivot_columns}' not found in file table"
        assert 'file' in files.columns.values, "Column 'file' must be present in the file table"
        assert isinstance(index, list), f"index must be a list (not {type(index)})"

        # Get the list of columns from the inputs
        input_columns = files.columns.values

        # Format as a wide dataset
        # Note that all the columns in `index` will be added if they are not already present
        wide_df = files.reindex(
            columns=index + pivot_columns + ['file']
        )
        wide_df = wide_df.pivot(
            index=index,
            columns=pivot_columns,
            values='file'
        )
        # Rename the columns to have a prefix, e.g. 'fastq_'
        wide_df = wide_df.rename(
            columns=lambda i: f"{column_prefix}{int(i)}"
        )
        wide_df = wide_df.reset_index()

        # Remove any columns from the output which were added from `index`
        for cname in index:
            if cname not in input_columns:
                wide_df = wide_df.drop(columns=[cname])
        # Remove any extra unnecessary columns
        wide_df = wide_df.drop(columns=pivot_columns, errors='ignore')
        return wide_df

    def wide_samplesheet(
            self,
            index=None,
            columns='read',
            values="file",  # noqa
            column_prefix="fastq_"
    ):
        """
        Format the samplesheet into a wide format with each sample on a row

        This is a legacy method, please use `pivot_samplesheet` instead.
        """
        warnings.warn("`wide_samplesheet` is deprecated, use `pivot_samplesheet` instead.",
                      DeprecationWarning, stacklevel=2)
        if values != "file":
            raise ValueError("The only supported value for `values` is 'file'")
        return self.pivot_files(index=index, pivot_columns=[columns], column_prefix=column_prefix)



from typing import List

from cirro_api_client.v1.models import ShareType

from cirro import CirroApi
from cirro.models.dataset import DatasetWithShare


def list_all_datasets(project_id: str, client: CirroApi) -> List[DatasetWithShare]:
    """
    List all datasets for a given project, including those shared with the project

    Args:
        project_id (str): ID of the Project
        client (cirro.CirroApi): Cirro API client

    Returns:
        List of datasets
    """
    datasets = client.datasets.list(project_id=project_id)
    datasets = [DatasetWithShare.from_dataset(d, share=None) for d in datasets]
    # Pull datasets from subscribed shares
    subscribed_shares = client.shares.list(project_id=project_id, share_type=ShareType.SUBSCRIBER)
    for share in subscribed_shares:
        datasets_in_share = client.datasets.list_shared(project_id, share.id)
        datasets_in_share = [DatasetWithShare.from_dataset(d, share=share) for d in datasets_in_share]
        datasets += datasets_in_share
    return datasets



from typing import List

from cirro_api_client.v1.models import Reference

from cirro.cirro_client import CirroApi
from cirro.models.file import File
from cirro.sdk.asset import DataPortalAssets, DataPortalAsset
from cirro.sdk.file import DataPortalFile


class DataPortalReference(DataPortalAsset):
    """
    Reference data object containing files which can be used for analysis in a particular project.
    """
    def __init__(self, ref: Reference, project_id: str, client: CirroApi):
        """
        Instantiate by listing the references which have been added to a particular project
        ```python
        from cirro import DataPortal()
        portal = DataPortal()
        project = portal.get_project_by_name("Project Name")
        references = project.list_references()
        ```
        """
        self._data = ref
        self._files = [
            DataPortalFile(File.from_file_entry(f, project_id), client) for f in ref.files
        ]

    @property
    def files(self) -> List[DataPortalFile]:
        """File(s) contained in the reference"""
        return self._files

    @property
    def name(self) -> str:
        """Reference name"""
        return self._data.name

    @property
    def type(self) -> str:
        """Type of reference data (e.g. genome_fasta)"""
        return self._data.type

    @property
    def absolute_path(self):
        if len(self._files) == 0:
            return None
        return self._files[0].absolute_path

    def __str__(self):
        return self.name


class DataPortalReferences(DataPortalAssets[DataPortalReference]):
    """Collection of DataPortalReference objects."""
    asset_name = "reference"

    def get_by_id(self, _id: str) -> DataPortalReference:
        raise NotImplementedError("Filtering by ID is not supported, use get_by_name")



class DataPortalAssetNotFound(Exception):
    """Exception raised when a Data Portal Asset cannot be found."""
    pass


class DataPortalInputError(Exception):
    """Exception raised invalid inputs are provided to the Data Portal."""
    pass



from abc import ABC, abstractmethod

from cirro_api_client.cirro_auth import AuthMethod


class AuthInfo(ABC):
    @abstractmethod
    def get_current_user(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def get_auth_method(self) -> AuthMethod:
        raise NotImplementedError()



from typing import Any, Optional
import logging

import WDL


def get_wdl_json_schema(wdl_file: str, logger: Optional[logging.Logger] = None) -> dict[str, Any]:  # type: ignore
    """
    Generate a JSON schema for parameters from a WDL workflow.

    :param wdl_file: Path to the main WDL file.
    """
    # Load the WDL document
    doc = WDL.load(wdl_file)

    # strictly look for top level workflow inputs
    if not doc.workflow:
        msg = "WDL Document does not contain a workflow."
        if logger:
            logger.error(msg)
        raise ValueError(msg)

    if not doc.workflow.inputs:
        msg = "WDL Document workflow does not contain inputs."
        if logger:
            logger.error(msg)
        raise ValueError(msg)

    # only get declarations from top-level workflow inputs
    params = {
        child.name: {
            'obj': child,
            'type': str(child.type),
            'default': child.expr.literal.value if child.expr and child.expr.literal else None,
            'optional': child.type.optional}
        for child in (doc.workflow.inputs or []) if isinstance(child, WDL.Decl)  # type: ignore
    }

    schema = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$id": "urn:cirro:wdl:schema",
        "type": "object",
        "properties": {},
        "required": []
    }

    type_map = {
        'String': 'string',
        'Int': 'integer',
        'Float': 'number',
        'Boolean': 'boolean',
        'File': 'string',  # File is typically a string path
        'Directory': 'string',  # Directory is typically a string path
        'Array[String]': 'array',
        'Array[Int]': 'array',
        'Array[Float]': 'array',
        'Array[Boolean]': 'array',
        'Array[File]': 'array',
        'Array[Directory]': 'array',

        # complex types
        'Pair': 'string',  # WDL Pair type, encode as JSON string
        'Map': 'string',  # WDL Map type, encode as JSON string
        'Struct': 'string',  # WDL Struct type, encode as JSON string

        # unsuported types
        # 'Object': 'object',  # WDL Object type (deprecated), not directly supported in JSON Schema
    }

    # convert params into JSON schema properties
    for name, param in params.items():
        if param['type'].replace('?', '') not in type_map:
            msg = f"Unsupported WDL type: {param['type']} for parameter {name}"
            if logger:
                logger.error(msg)
            raise ValueError(msg)

        schema['properties'][name] = {
            "type": type_map[param['type'].replace('?', '')],  # remove optional marker
            "wdlType": param['type'],
            "default": param['default'],
            "description": f"Parameter {name} of type {param['type']}"
        }
        # paramters that are required have default=None and optional=False
        if param['default'] is None and not param['optional']:
            schema['required'].append(name)

    return schema



from pathlib import Path
from typing import List, Union, Callable, TypeVar, Optional

from prompt_toolkit.validation import Validator, ValidationError
import questionary
from questionary import prompt


class InputError(Exception):
    """Errors detected from user input"""
    pass


class DirectoryValidator(Validator):
    def validate(self, document):
        is_a_directory = Path(document.text).expanduser().is_dir()
        if not is_a_directory or len(document.text.strip()) == 0:
            raise ValidationError(
                message='Please enter a valid directory',
                cursor_position=len(document.text)
            )


def prompt_wrapper(questions):
    answers = prompt(questions)
    # Prompt catches KeyboardInterrupt and sends back an empty dictionary
    # We want to catch this exception
    if len(answers) == 0:
        raise KeyboardInterrupt()
    return answers


def type_validator(t, v):
    """Return a boolean indicating whether `v` can be cast to `t(v)` without raising a ValueError."""
    try:
        t(v)
        return True
    except ValueError:
        return False


def ask(function_name: str,
        msg: str,
        validate_type=None,
        output_transformer: Callable = None,
        **kwargs) -> Union[str, List[str]]:
    """
    Wrap questionary functions to catch escapes and exit gracefully.
    function_name: https://questionary.readthedocs.io/en/stable/pages/types.html#
    """

    # Get the questionary function
    questionary_f = questionary.__dict__.get(function_name)

    # Make sure that the function exists
    assert questionary_f is not None, f"No such questionary function: {function_name}"

    if kwargs.get("use_shortcuts") is None and function_name == "select":
        kwargs["use_shortcuts"] = True

    if validate_type is not None:
        kwargs["validate"] = lambda v: type_validator(validate_type, v)

    # The default value must be a string
    if kwargs.get("default") is not None:
        kwargs["default"] = str(kwargs["default"])

    if kwargs.get("required"):
        del kwargs["required"]
        kwargs["validate"] = lambda val: len(val.strip()) > 0 or "This field is required"

    # Add a spacer line before asking the question
    print("")

    # Get the response
    resp = questionary_f(msg, **kwargs).ask()

    # If the user escaped the question
    if resp is None:
        raise KeyboardInterrupt()

    # If an output transformation function was defined
    if output_transformer is not None:
        # Call the function
        resp = output_transformer(resp)

    # Otherwise
    return resp


def ask_yes_no(msg):
    return ask("select", msg, choices=["Yes", "No"]) == "Yes"


T = TypeVar('T')


def get_id_from_name(items: List[T], name_or_id: str) -> Optional[str]:
    matched = get_item_from_name_or_id(items, name_or_id)
    if not matched:
        item_type = type(items[0]).__name__
        raise InputError(f"Could not find {item_type} {name_or_id}")
    return matched.id


def get_item_from_name_or_id(items: List[T], name_or_id: str) -> Optional[T]:
    matched = next((p for p in items if p.id == name_or_id), None)
    if matched:
        return matched
    return next((p for p in items if p.name == name_or_id), None)



import os
from os import path
from functools import cached_property
from typing import Any, Iterable, Optional
import logging
from enum import Enum

from referencing import Resource, Registry


CONFIG_APP_URL = "https://app.cirro.bio/tools/pipeline-configurator/"


class ConfigAppStatus(Enum):
    """
    Enum to represent the status of the config app recommendation.
    """
    RECOMMENDED = "recommended"
    OPTIONAL = "optional"


class PipelineDefinition:
    """
    A pipeline definition on disk.
    """

    def __init__(self, root_dir: str, entrypoint: Optional[str] = None, logger: Optional[logging.Logger] = None):
        self.root_dir: str = path.expanduser(path.expandvars(root_dir))
        self.entrypoint: Optional[str] = entrypoint
        self.config_app_status: ConfigAppStatus = ConfigAppStatus.RECOMMENDED

        if logger:
            self.logger: logging.Logger = logger
        else:
            log_formatter = logging.Formatter(
                '%(asctime)s %(levelname)-8s [PipelineDefinition] %(message)s'
            )
            self.logger = logging.getLogger()
            self.logger.setLevel(logging.INFO)
            console_handler = logging.StreamHandler()
            console_handler.setFormatter(log_formatter)
            self.logger.addHandler(console_handler)

    @cached_property
    def parameter_schema(self) -> Resource:
        """
        Returns the parameter schema for the pipeline.
        """
        workflow_files = os.walk(self.root_dir, topdown=True)
        for dirpath, dirnames, filenames in workflow_files:
            # look for a nextflow_schema.json file at the root of the workflow directory
            is_nextflow = (
                ('nextflow_schema.json' in filenames)
                or ('nextflow.config' in filenames)
            )
            if is_nextflow:
                # lazy load nextflow dependencies
                from cirro.models.process.nextflow import get_nextflow_json_schema
                contents = get_nextflow_json_schema(self.root_dir, self.logger)

                if 'nextflow_schema.json' in filenames:
                    self.config_app_status = ConfigAppStatus.OPTIONAL

                break
            elif any(f.endswith('.wdl') for f in filenames):  # is WDL
                # generate schema from WDL workflow
                wdl_file = None
                if self.entrypoint:
                    self.logger.info(f"Using entrypoint WDL file: {self.entrypoint}")
                    # if an entrypoint is specified, look for that specific WDL file
                    wdl_file = next((f for f in filenames if f.endswith(self.entrypoint)), None)
                    if not wdl_file:
                        raise FileNotFoundError(f"Entrypoint WDL file '{self.entrypoint}' not found in {dirpath}")
                else:
                    # otherwise, just take the first WDL file found
                    wdl_file = next(f for f in filenames if f.endswith('.wdl'))

                # lazy load wdl dependencies
                from cirro.models.process.wdl import get_wdl_json_schema

                wdl_file = path.join(dirpath, wdl_file)
                contents = get_wdl_json_schema(wdl_file, self.logger)
                break

        else:
            raise RuntimeError("Unrecognized workflow format. Please provide a valid Nextflow or WDL workflow.")

        schema = Resource.from_contents(contents)
        return schema

    @property
    def form_configuration(self) -> dict[str, Any]:
        """
        Returns the form configuration for the pipeline.
        """
        return {
            "form": self.parameter_schema.contents,
            "ui": {}
        }

    @property
    def input_configuration(self) -> dict[str, str]:
        """
        Returns the input configuration for the pipeline.
        """
        schema = self.parameter_schema
        contents = schema.contents

        parameters = {
            p["name"]: p["jsonPath"]
            for p in get_input_params('$.dataset.params', contents, schema)
        }
        return parameters

    def __repr__(self):
        return f"PipelineDefinition(name={self.root_dir})"


def get_input_params(property_path: str, definition: dict[str, Any], schema: Resource) -> Iterable[dict[str, Any]]:
    resolved = definition
    if '$ref' in definition:
        registry = schema @ Registry()
        resolver = registry.resolver()

        resolved = resolver.lookup(f"{schema.id()}{definition['$ref']}").contents

    if resolved.get('type') == 'object':
        # recursively get input params for nested objects
        for p, d in resolved.get('properties', {}).items():
            nested_path = f"{property_path}.{p}"
            yield from get_input_params(nested_path, d, schema)
    else:
        json_path = property_path
        param_is_path = (
            (resolved.get('wdlType') and resolved['wdlType'].replace('?', '') in ('File', 'Directory'))
            or (resolved.get('format') and resolved['format'] in ('file-path', 'directory-path'))
        )
        if param_is_path:
            # override the jsonPath to be '$.inputs[*].dataPath'
            json_path = "$.inputs[*].dataPath"

        yield {
            'name': property_path.split('.')[-1],
            'type': resolved.get('type', 'string'),
            'default': resolved.get('default', None),
            'jsonPath': json_path,
        }



from pathlib import Path

from cirro_api_client.v1.models import Project, ReferenceType, ReferenceTypeValidationItem
from prompt_toolkit.validation import Validator, ValidationError

from cirro.cli.interactive.common_args import ask_project
from cirro.cli.interactive.utils import ask
from cirro.cli.models import UploadReferenceArguments
from cirro.helpers.references import get_matching_validation, format_expected_file


class ReferenceFileValidator(Validator):
    def __init__(self, validation: ReferenceTypeValidationItem):
        self.validation = validation

    def validate(self, document):
        file_path = Path(document.text)
        if not file_path.is_file() or len(document.text.strip()) == 0:
            raise ValidationError(
                message='Please enter a valid file',
                cursor_position=len(document.text)
            )
        matching = get_matching_validation(file_path.name, [self.validation])
        if not matching:
            raise ValidationError(
                message=f'File "{file_path.name}" does not match expected file: '
                        f'{format_expected_file(self.validation)}',
                cursor_position=len(document.text)
            )


def ask_reference_type(reference_types: list[ReferenceType], input_value: str = None) -> ReferenceType:
    reference_type_names = sorted([ref_type.name for ref_type in reference_types])
    reference_type_name = ask(
        'autocomplete',
        'What is the type of reference? (use TAB to display options)',
        choices=reference_type_names,
        default=input_value if input_value in reference_type_names else '',
        validate=lambda v: v.strip() != '' or 'This field is required'
    )
    for ref_type in reference_types:
        if ref_type.name == reference_type_name:
            return ref_type
    raise ValueError(f'Reference type "{reference_type_name}" not found')


def ask_name(input_value: str = None) -> str:
    return ask(
        'text',
        'What is the name of the reference?',
        default=input_value if input_value else '',
        validate=lambda v: v.strip() != '' or 'This field is required'
    )


def ask_reference_file_single(validation: ReferenceTypeValidationItem) -> Path:
    file_path = ask(
        'path',
        f'Please provide the path to {format_expected_file(validation)}',
        validate=ReferenceFileValidator(validation)
    )
    return Path(file_path)


def ask_reference_files(reference_type: ReferenceType) -> list[Path]:
    reference_files = []
    for validation in reference_type.validation:
        file_path = ask_reference_file_single(validation)
        reference_files.append(Path(file_path))
    return reference_files


def gather_reference_upload_arguments(input_params: UploadReferenceArguments,
                                      projects: list[Project],
                                      reference_types: list[ReferenceType]):
    input_params['project'] = ask_project(projects, input_params.get('project'))
    input_params['name'] = ask_name(input_params.get('name'))

    reference_type = ask_reference_type(reference_types, input_params.get('reference_type'))
    print('This reference type expects the following files:')
    for validation in reference_type.validation:
        print('  - ' + format_expected_file(validation))
    input_params['reference_type'] = reference_type.name

    files = ask_reference_files(reference_type)
    if len(files) == 0:
        raise ValueError("At least one reference file must be provided.")
    return input_params, files



import requests

from cirro.models.file import FileAccessContext
from cirro.services import FileService


def _get_file_from_path(self: FileService, access_context: FileAccessContext, file_path: str):
    # Load lazily -- must be installed with the pyodide extras
    try:
        from requests_aws4auth import AWS4Auth # noqa
    except ModuleNotFoundError:
        raise ModuleNotFoundError("Could not load requests_aws4auth -- make sure to install cirro[pyodide]")

    creds = self.get_access_credentials(access_context)

    # Construct the URL endpoint to the file
    endpoint = "/".join([
        f"https://{access_context.bucket}.s3.{creds.region}.amazonaws.com",
        access_context.prefix,
        file_path
    ])

    # Set up the authorization
    auth = AWS4Auth(
        creds.access_key_id,
        creds.secret_access_key,
        creds.region,
        's3',
        session_token=creds.session_token,
    )

    # Make the request
    response = requests.get(endpoint, auth=auth, stream=True)

    # Raise an error if the exit code is not 400
    response.raise_for_status()

    # Read the content of the response and decode it
    content = response.raw.read(decode_content=True)

    # Return the contents of the response
    return content


def pyodide_patch_file():
    """
    Patch the FileService to use requests to get files from S3,
     rather than S3Client (boto3)
    """
    FileService.get_file_from_path = _get_file_from_path



from typing import List, Optional, Dict

from cirro_api_client.v1.api.execution import run_analysis, stop_analysis, get_project_summary, \
    get_tasks_for_execution, get_task_logs, get_execution_logs
from cirro_api_client.v1.api.processes import get_process_parameters
from cirro_api_client.v1.models import RunAnalysisRequest, CreateResponse, Task

from cirro.models.form_specification import ParameterSpecification
from cirro.services.base import BaseService


class ExecutionService(BaseService):
    """
    Service for interacting with the Execution endpoints
    """
    def run_analysis(self, project_id: str, request: RunAnalysisRequest) -> CreateResponse:
        """
        Launch an analysis job running a process on a set of inputs

        Args:
            project_id (str): ID of the Project
            request (cirro_api_client.v1.models.RunAnalysisRequest):

        Returns:
            The ID of the created dataset

        ```python
        from cirro_api_client.v1.models import RunAnalysisRequest, RunAnalysisRequestParams
        from cirro.cirro_client import CirroApi

        # Example:
        # Run the "process-nf-core-rnaseq-3_8" process using input data
        # from a dataset with the id "source-dataset-id"

        # Optional analysis parameters
        params = RunAnalysisRequestParams.from_dict({
            "param_a": "val_a",
            "param_b": "val_b"
        })

        cirro = CirroApi()
        request = RunAnalysisRequest(
            name="Name of the newly created dataset",
            description="Longer description of newly created dataset",
            process_id="process-nf-core-rnaseq-3_8",
            source_dataset_ids=["source-dataset-id"],
            params=params
        )
        cirro.execution.run_analysis("project-id", request)
        ```
        """

        form_spec = get_process_parameters.sync(
            process_id=request.process_id,
            client=self._api_client
        )

        ParameterSpecification(
            form_spec
        ).validate_params(
            request.params.to_dict() if request.params else {}
        )

        return run_analysis.sync(
            project_id=project_id,
            body=request,
            client=self._api_client
        )

    def stop_analysis(self, project_id: str, dataset_id: str):
        """
        Terminates all jobs related to a running analysis

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
        """

        return stop_analysis.sync(
            project_id=project_id,
            dataset_id=dataset_id,
            client=self._api_client
        )

    def get_project_summary(self, project_id: str) -> Dict[str, List[Task]]:
        """
        Gets an overview of the executions currently running in the project, by job queue

        Args:
            project_id (str): ID of the Project

        Returns:
            `cirro_api_client.v1.models.GetProjectSummaryResponse200`
        """

        return get_project_summary.sync(
            project_id=project_id,
            client=self._api_client
        ).additional_properties

    def get_execution_logs(self, project_id: str, dataset_id: str, force_live=False) -> str:
        """
        Gets live logs from main execution task

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
            force_live (bool): If True, it will fetch logs from CloudWatch,
                even if the execution is already completed

        """

        resp = get_execution_logs.sync(
            project_id=project_id,
            dataset_id=dataset_id,
            force_live=force_live,
            client=self._api_client
        )

        return '\n'.join(e.message for e in resp.events)

    def get_tasks_for_execution(self, project_id: str, dataset_id: str, force_live=False) -> Optional[List[Task]]:
        """
        Gets the tasks submitted by the workflow execution

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
            force_live (bool): If True, it will try to get the list of jobs
                from the executor (i.e., AWS Batch), rather than the workflow report
        """

        return get_tasks_for_execution.sync(
            project_id=project_id,
            dataset_id=dataset_id,
            force_live=force_live,
            client=self._api_client
        )

    def get_task_logs(self, project_id: str, dataset_id: str, task_id: str, force_live=False) -> str:
        """
        Gets the log output from an individual task

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
            task_id (str): ID of the task
            force_live (bool): If True, it will fetch logs from CloudWatch,
                even if the execution is already completed

        """

        resp = get_task_logs.sync(
            project_id=project_id,
            dataset_id=dataset_id,
            task_id=task_id,
            force_live=force_live,
            client=self._api_client
        )

        return '\n'.join(e.message for e in resp.events)



import datetime
from typing import Union, List, Optional

from cirro_api_client.v1.models import Dataset, DatasetDetail, RunAnalysisRequest, ProcessDetail, Status, \
    RunAnalysisRequestParams, Tag, ArtifactType, NamedItem

from cirro.cirro_client import CirroApi
from cirro.models.assets import DatasetAssets
from cirro.sdk.asset import DataPortalAssets, DataPortalAsset
from cirro.sdk.exceptions import DataPortalAssetNotFound
from cirro.sdk.exceptions import DataPortalInputError
from cirro.sdk.file import DataPortalFile, DataPortalFiles
from cirro.sdk.helpers import parse_process_name_or_id
from cirro.sdk.process import DataPortalProcess


class DataPortalDataset(DataPortalAsset):
    """
    Datasets in the Data Portal are collections of files which have
    either been uploaded directly, or which have been output by
    an analysis pipeline or notebook.
    """

    def __init__(self, dataset: Union[Dataset, DatasetDetail], client: CirroApi):
        """
        Instantiate a dataset object

        Should be invoked from a top-level constructor, for example:

        ```python
        from cirro import DataPortal()
        portal = DataPortal()
        dataset = portal.get_dataset(
            project="id-or-name-of-project",
            dataset="id-or-name-of-dataset"
        )
        ```

        """
        assert dataset.project_id is not None, "Must provide dataset with project_id attribute"
        self._data = dataset
        self._assets: Optional[DatasetAssets] = None
        self._client = client

    @property
    def id(self) -> str:
        """Unique identifier for the dataset"""
        return self._data.id

    @property
    def name(self) -> str:
        """Editable name for the dataset"""
        return self._data.name

    @property
    def description(self) -> str:
        """Longer name for the dataset"""
        return self._data.description

    @property
    def process_id(self) -> str:
        """Unique ID of process used to create the dataset"""
        return self._data.process_id

    @property
    def process(self) -> ProcessDetail:
        """
        Object representing the process used to create the dataset
        """
        return self._client.processes.get(self.process_id)

    @property
    def project_id(self) -> str:
        """ID of the project containing the dataset"""
        return self._data.project_id

    @property
    def status(self) -> Status:
        """
        Status of the dataset
        """
        return self._data.status

    @property
    def source_dataset_ids(self) -> List[str]:
        """IDs of the datasets used as sources for this dataset (if any)"""
        return self._data.source_dataset_ids

    @property
    def source_datasets(self) -> List['DataPortalDataset']:
        """
        Objects representing the datasets used as sources for this dataset (if any)
        """
        return [
            DataPortalDataset(
                dataset=self._client.datasets.get(project_id=self.project_id, dataset_id=dataset_id),
                client=self._client
            )
            for dataset_id in self.source_dataset_ids
        ]

    @property
    def params(self) -> dict:
        """
        Parameters used to generate the dataset
        """
        return self._get_detail().params.to_dict()

    @property
    def info(self) -> dict:
        """
        Extra information about the dataset
        """
        return self._get_detail().info.to_dict()

    @property
    def tags(self) -> List[Tag]:
        """
        Tags applied to the dataset
        """
        return self._data.tags

    @property
    def share(self) -> Optional[NamedItem]:
        """
        Share associated with the dataset, if any.
        """
        return self._get_detail().share

    @property
    def created_by(self) -> str:
        """User who created the dataset"""
        return self._data.created_by

    @property
    def created_at(self) -> datetime.datetime:
        """Timestamp of dataset creation"""
        return self._data.created_at

    def _get_detail(self):
        if not isinstance(self._data, DatasetDetail):
            self._data = self._client.datasets.get(project_id=self.project_id, dataset_id=self.id)
        return self._data

    def _get_assets(self):
        if not self._assets:
            self._assets = self._client.datasets.get_assets_listing(
                project_id=self.project_id,
                dataset_id=self.id
            )
        return self._assets

    def __str__(self):
        return '\n'.join([
            f"{i.title()}: {self.__getattribute__(i)}"
            for i in ['name', 'id', 'description', 'status']
        ])

    def get_file(self, relative_path: str) -> DataPortalFile:
        """
        Get a file from the dataset using its relative path.

        Args:
            relative_path (str): Relative path of file within the dataset

        Returns:
            `from cirro.sdk.file import DataPortalFile`
        """

        # Get the list of files in this dataset
        files = self.list_files()

        # Try getting the file using the relative path provided by the user
        try:
            return files.get_by_id(relative_path)
        except DataPortalAssetNotFound:
            # Try getting the file with the 'data/' prefix prepended
            try:
                return files.get_by_id("data/" + relative_path)
            except DataPortalAssetNotFound:
                # If not found, raise the exception using the string provided
                # by the user, not the data/ prepended version (which may be
                # confusing to the user)
                msg = '\n'.join([f"No file found with path '{relative_path}'."])
                raise DataPortalAssetNotFound(msg)

    def list_files(self) -> DataPortalFiles:
        """
        Return the list of files which make up the dataset.
        """
        files = self._get_assets().files
        return DataPortalFiles(
            [
                DataPortalFile(file=file, client=self._client)
                for file in files
            ]
        )

    def get_artifact(self, artifact_type: ArtifactType) -> DataPortalFile:
        """
        Get the artifact of a particular type from the dataset
        """
        artifacts = self._get_assets().artifacts
        artifact = next((a for a in artifacts if a.artifact_type == artifact_type), None)
        if artifact is None:
            raise DataPortalAssetNotFound(f"No artifact found with type '{artifact_type}'")
        return DataPortalFile(file=artifact.file, client=self._client)

    def list_artifacts(self) -> List[DataPortalFile]:
        """
        Return the list of artifacts associated with the dataset

        An artifact may be something generated as part of the analysis or other process.
        See `cirro_api_client.v1.models.ArtifactType` for the list of possible artifact types.

        """
        artifacts = self._get_assets().artifacts
        return DataPortalFiles(
            [
                DataPortalFile(file=artifact.file, client=self._client)
                for artifact in artifacts
            ]
        )

    def download_files(self, download_location: str = None) -> None:
        """
        Download all the files from the dataset to a local directory.

        Args:
            download_location (str): Path to local directory
        """

        # Alias for internal method
        self.list_files().download(download_location)

    def run_analysis(
            self,
            name: str = None,
            description: str = "",
            process: Union[DataPortalProcess, str] = None,
            params=None,
            notifications_emails: List[str] = None,
            compute_environment: str = None,
            resume_dataset_id: str = None
    ) -> str:
        """
        Runs an analysis on a dataset, returns the ID of the newly created dataset.

        The process can be provided as either a DataPortalProcess object,
        or a string which corresponds to the name or ID of the process.

        Args:
            name (str): Name of newly created dataset
            description (str): Description of newly created dataset
            process (DataPortalProcess or str): Process to run
            params (dict): Analysis parameters
            notifications_emails (List[str]): Notification email address(es)
            compute_environment (str): Name or ID of compute environment to use,
             if blank it will run in AWS
            resume_dataset_id (str): ID of dataset to resume from, used for caching task execution.
             It will attempt to re-use the previous output to minimize duplicate work

        Returns:
            dataset_id (str): ID of newly created dataset
        """
        if name is None:
            raise DataPortalInputError("Must specify 'name' for run_analysis")
        if process is None:
            raise DataPortalInputError("Must specify 'process' for run_analysis")
        if notifications_emails is None:
            notifications_emails = []
        if params is None:
            params = {}

        # If the process is a string, try to parse it as a process name or ID
        process = parse_process_name_or_id(process, self._client)

        if compute_environment:
            compute_environments = self._client.compute_environments.list_environments_for_project(
                project_id=self.project_id
            )
            compute_environment = next(
                (env for env in compute_environments
                 if env.name == compute_environment or env.id == compute_environment),
                None
            )
            if compute_environment is None:
                raise DataPortalInputError(f"Compute environment '{compute_environment}' not found")

        resp = self._client.execution.run_analysis(
            project_id=self.project_id,
            request=RunAnalysisRequest(
                name=name,
                description=description,
                process_id=process.id,
                source_dataset_ids=[self.id],
                params=RunAnalysisRequestParams.from_dict(params),
                notification_emails=notifications_emails,
                resume_dataset_id=resume_dataset_id,
                compute_environment_id=compute_environment.id if compute_environment else None
            )
        )
        return resp.id


class DataPortalDatasets(DataPortalAssets[DataPortalDataset]):
    """Collection of multiple DataPortalDataset objects."""
    asset_name = "dataset"



from pathlib import Path
from typing import Optional

from prompt_toolkit.shortcuts import CompleteStyle

from cirro.cli.interactive import utils


def ask_pipeline_root_directory(input_value: str) -> str:
    """
    Asks the user for the root directory of the pipeline definition.
    """
    root_dir_prompt = {
        'type': 'path',
        'name': 'pipeline_dir',
        'message': 'Enter the root directory for the pipeline definition',
        'validate': utils.DirectoryValidator,
        'default': input_value or '',
        'complete_style': CompleteStyle.READLINE_LIKE,
        'only_directories': True
    }

    answers = utils.prompt_wrapper(root_dir_prompt)
    return str(Path(answers['pipeline_dir']).expanduser())


def ask_pipeline_entrypoint() -> Optional[str]:
    """
    Asks the user for the entrypoint script of the pipeline.
    """
    entrypoint_prompt = {
        'type': 'input',
        'name': 'entrypoint',
        'message': 'Enter the entrypoint script (optional, WDL only)',
    }

    answers = utils.prompt_wrapper(entrypoint_prompt)
    return answers['entrypoint'].strip() if answers['entrypoint'] else None


def ask_output_directory(input_value: str) -> str:
    """
    Asks the user for the output directory to write pipeline configuration files.
    """
    output_dir_prompt = {
        'type': 'path',
        'name': 'output_dir',
        'message': 'Enter the output directory for the pipeline configuration files',
        'default': input_value or str(Path.cwd()),
        'complete_style': CompleteStyle.READLINE_LIKE,
        'only_directories': True
    }

    answers = utils.prompt_wrapper(output_dir_prompt)
    return str(Path(answers['output_dir']).expanduser())


def gather_create_pipeline_config_arguments(input_params):
    """
    Gathers input parameters for creating pipeline configuration files interactively.
    """
    input_params['pipeline_dir'] = ask_pipeline_root_directory(input_params.get('pipeline_dir'))
    input_params['entrypoint'] = ask_pipeline_entrypoint()
    input_params['output_dir'] = ask_output_directory(input_params.get('output_dir'))
    return input_params



import os
from os import path
import json
from typing import Any, Optional
import logging

from nf_core.pipelines.schema import PipelineSchema


def get_nextflow_json_schema(workflow_dir: str, logger: Optional[logging.Logger] = None) -> dict[str, Any]:
    """
    Get or generate a JSON schema for parameters from a Nextflow workflow.

    :param workflow_dir: Workflow definition directory.
    """

    workflow_files = os.walk(workflow_dir, topdown=True, followlinks=True)
    for dirpath, dirnames, filenames in workflow_files:
        if 'nextflow_schema.json' in filenames:
            # check if a nextflow_schema.json file exists at the root of the workflow directory
            schema_path = path.join(dirpath, 'nextflow_schema.json')
            if logger:
                logger.info(f"Nextflow schema found at {schema_path}")
            with open(schema_path, 'r') as f:
                contents = json.load(f)
            break
        else:
            # generate the parameter schema non-interactively
            ps = PipelineSchema()

            # since this is already filtered to a case where a nextflow_schema.json file is not found,
            # only use a subset of functionality from the build_schema method and use our logger if provided.
            try:
                ps.get_schema_path(workflow_dir, local_only=True)
            except AssertionError:
                # if the schema path is not found, we will generate it
                if logger:
                    logger.info("Nextflow schema path not found, generating schema from workflow files.")

            ps.get_wf_params()
            ps.make_skeleton_schema()
            ps.remove_schema_notfound_configs()
            ps.remove_schema_empty_definitions()
            ps.add_schema_found_configs()

            try:
                ps.validate_schema()
            except AssertionError as e:
                if logger:
                    logger.error(f"Nextflow schema creation failed: {e}")
                raise e

            contents = ps.schema
            break
    else:
        msg = "Nextflow schema not found or could not be generated."
        if logger:
            logger.error(msg)
        raise FileNotFoundError(msg)

    # ignore the 'allOf' attribute if it exists
    # merge 'definitions' or '$defs' attribute with 'properties'
    if contents.get('allOf'):
        del contents['allOf']

    contents['properties'] = (
        contents.get('properties', {})
        | contents.get('$defs', {})
        | contents.get('definitions', {})
    )

    return contents



from cirro.helpers.form import FormBuilder
from cirro.helpers.preprocess_dataset import PreprocessDataset
from cirro.helpers.pyodide_patch_file import pyodide_patch_file
from cirro.helpers.pyodide_patch_httpx import pyodide_patch_httpx

__all__ = [
    'PreprocessDataset',
    'FormBuilder'
]


def pyodide_patch_all():
    pyodide_patch_file()
    pyodide_patch_httpx()



from cirro.auth import DeviceCodeAuth
from cirro.cirro_client import CirroApi
from cirro.config import AppConfig
from cirro.sdk.portal import DataPortal


class DataPortalLogin:
    """
    Start the login process, obtaining the authorization message from Cirro
    needed to confirm the user identity.

    Useful when you need to authenticate a user in a non-blocking way.

    Usage:

    ```python
    # Replace app.cirro.bio as appropriate
    login = DataPortalLogin(base_url="app.cirro.bio")

    # Present the user with the authorization message
    print(login.auth_message)

    # Generate the authenticated DataPortal object,
    # blocking until the user completes the login process in their browser
    portal = login.await_completion()
    ```
    """
    base_url: str
    auth_info: DeviceCodeAuth

    def __init__(self, base_url: str = None, enable_cache=False):
        app_config = AppConfig(base_url=base_url)

        self.base_url = base_url

        self.auth_info = DeviceCodeAuth(
            region=app_config.region,
            client_id=app_config.client_id,
            auth_endpoint=app_config.auth_endpoint,
            enable_cache=enable_cache,
            await_completion=False
        )

    @property
    def auth_message(self) -> str:
        """Authorization message provided by Cirro."""
        return self.auth_info.auth_message

    @property
    def auth_message_markdown(self) -> str:
        """Authorization message provided by Cirro (Markdown format)."""
        return self.auth_info.auth_message_markdown

    def await_completion(self) -> DataPortal:
        """Complete the login process and return an authenticated client"""

        # Block until the user completes the login flow
        self.auth_info.await_completion()

        # Set up the client object
        cirro_client = CirroApi(
            auth_info=self.auth_info,
            base_url=self.base_url
        )

        # Return the Data Portal object
        return DataPortal(client=cirro_client)



import cirro.file_utils  # noqa
from cirro.cirro_client import CirroApi
from cirro.sdk.dataset import DataPortalDataset
from cirro.sdk.login import DataPortalLogin
from cirro.sdk.portal import DataPortal
from cirro.sdk.process import DataPortalProcess
from cirro.sdk.project import DataPortalProject
from cirro.sdk.reference import DataPortalReference

__all__ = [
    'DataPortal',
    'DataPortalLogin',
    'DataPortalProject',
    'DataPortalProcess',
    'DataPortalDataset',
    'DataPortalReference',
    'CirroApi',
    'file_utils'
]



from attrs import define as _attrs_define
from cirro_api_client.v1.models import Share, Dataset


@_attrs_define
class DatasetWithShare(Dataset):
    share: Share

    @classmethod
    def from_dataset(cls, dataset: Dataset, share: Share) -> 'DatasetWithShare':
        return cls(
            id=dataset.id,
            name=dataset.name,
            description=dataset.description,
            project_id=dataset.project_id,
            process_id=dataset.process_id,
            source_dataset_ids=dataset.source_dataset_ids,
            status=dataset.status,
            tags=dataset.tags,
            created_by=dataset.created_by,
            created_at=dataset.created_at,
            updated_at=dataset.updated_at,
            share=share
        )



from typing import List

from cirro_api_client.v1.api.sharing import get_shares, create_share, subscribe_share, unsubscribe_share, get_share, \
    update_share, delete_share
from cirro_api_client.v1.models import Share, ShareDetail, CreateResponse, ShareInput, ShareType

from cirro.services.base import BaseService


class ShareService(BaseService):
    """
    Service for interacting with the Share endpoints
    """

    def list(self, project_id: str, share_type: ShareType = None) -> List[Share]:
        """
        Retrieve a list of shares for a given project
        optionally filtered by share type

        Args:
            project_id (str): ID of the Project
            share_type (ShareType): Type of share to filter by

        ```python

        from cirro_api_client.v1.models import ShareType
        from cirro.cirro_client import CirroApi

        cirro = CirroApi()

        # List shares that we've published
        cirro.shares.list(project_id="project-id", share_type=ShareType.PUBLISHER)
        ```
        """
        shares = get_shares.sync(project_id=project_id, client=self._api_client)
        if share_type:
            shares = [share for share in shares if share.share_type == share_type]
        return shares

    def get(self, project_id: str, share_id: str) -> ShareDetail:
        """
        Get details of a share
        """
        return get_share.sync(project_id=project_id, share_id=share_id, client=self._api_client)

    def create(self, project_id: str, share: ShareInput) -> CreateResponse:
        """
        Publish a share for a given project
        """
        return create_share.sync(project_id=project_id,
                                 body=share,
                                 client=self._api_client)

    def update(self, project_id: str, share_id: str, share: ShareInput):
        """
        Publish a share for a given project
        """
        update_share.sync_detailed(project_id=project_id,
                                   share_id=share_id,
                                   body=share,
                                   client=self._api_client)

    def delete(self, project_id: str, share_id: str):
        """
        Delete a share
        """
        delete_share.sync_detailed(project_id=project_id,
                                   share_id=share_id,
                                   client=self._api_client)

    def subscribe(self, project_id: str, share_id: str):
        """
        Subscribe to a share for a given project
        """
        subscribe_share.sync_detailed(project_id=project_id,
                                      share_id=share_id,
                                      client=self._api_client)

    def unsubscribe(self, project_id: str, share_id: str):
        """
        Unsubscribe from a share for a given project
        """
        unsubscribe_share.sync_detailed(project_id=project_id,
                                        share_id=share_id,
                                        client=self._api_client)



from pathlib import Path
from typing import List, Optional

from cirro_api_client.v1.api.references import get_reference_types, get_references_for_project, \
    refresh_project_references
from cirro_api_client.v1.models import ReferenceType, Reference

from cirro.helpers.references import generate_reference_file_path_map
from cirro.models.file import PathLike, FileAccessContext
from cirro.services.file import FileEnabledService


class ReferenceService(FileEnabledService):
    """
    Service for interacting with the References endpoints
    """
    def get_types(self) -> Optional[List[ReferenceType]]:
        """
        List available reference types
        """
        return get_reference_types.sync(client=self._api_client)

    def get_type(self, reference_type_name: str) -> Optional[ReferenceType]:
        """
        Get a specific reference type by name

        Args:
            reference_type_name (str): Name of the reference type (e.g. "Reference Genome (FASTA)")
        """
        types = self.get_types()
        return next((t for t in types if t.name == reference_type_name), None)

    def get_for_project(self, project_id: str) -> Optional[List[Reference]]:
        """
        List available references for a given project
        """
        return get_references_for_project.sync(project_id=project_id, client=self._api_client)

    def upload_reference(self,
                         name: str,
                         reference_files: list[PathLike],
                         project_id: str,
                         ref_type: ReferenceType):
        """
        Upload a reference to a project

        Args:
            name (str): Name of the reference (e.g. "GRCh38")
                no spaces are allowed in the name
            reference_files (list[PathLike]): Path to the reference file to upload
            project_id (str): ID of the project to upload the reference to
            ref_type (ReferenceType): Type of the reference

        ```python
        from pathlib import Path

        from cirro import CirroApi

        cirro = CirroApi()

        crispr_library = cirro.references.get_type("CRISPR sgRNA Library")
        files = [
            Path("~/crispr_library.csv").expanduser()
        ]

        cirro.references.upload_reference(
            name="Library Name",
            project_id="project-id",
            ref_type=crispr_library,
            reference_files=files,
        )
        ```
        """
        # Validate name
        if ' ' in name:
            raise ValueError("Reference name cannot contain spaces")

        access_context = FileAccessContext.upload_reference(
            project_id=project_id,
            base_url=f's3://project-{project_id}/resources/data/references',
        )

        file_path_map = generate_reference_file_path_map(
            files=reference_files,
            name=name,
            ref_type=ref_type
        )

        self._file_service.upload_files(
            access_context=access_context,
            directory=Path('~').expanduser(),  # Full path expected in reference_files
            files=reference_files,
            file_path_map=file_path_map
        )

        refresh_project_references.sync_detailed(project_id=project_id,
                                                 client=self._api_client)



import importlib.metadata
import json
import logging
import os
import sys
from pathlib import Path

import requests
from cirro_api_client.v1.models import UploadDatasetRequest, Status, Executor

from cirro.cirro_client import CirroApi
from cirro.cli.interactive.auth_args import gather_auth_config
from cirro.cli.interactive.create_pipeline_config import gather_create_pipeline_config_arguments
from cirro.cli.interactive.download_args import gather_download_arguments, ask_dataset_files
from cirro.cli.interactive.download_args import gather_download_arguments_dataset
from cirro.cli.interactive.list_dataset_args import gather_list_arguments
from cirro.cli.interactive.upload_args import gather_upload_arguments
from cirro.cli.interactive.upload_reference_args import gather_reference_upload_arguments
from cirro.cli.interactive.utils import get_id_from_name, get_item_from_name_or_id, InputError
from cirro.cli.models import ListArguments, UploadArguments, DownloadArguments, CreatePipelineConfigArguments, \
    UploadReferenceArguments
from cirro.config import UserConfig, save_user_config, load_user_config
from cirro.file_utils import get_files_in_directory
from cirro.models.process import PipelineDefinition, ConfigAppStatus, CONFIG_APP_URL
from cirro.services.service_helpers import list_all_datasets

NO_PROJECTS = "No projects available"
# Log to STDOUT
log_formatter = logging.Formatter(
    '%(asctime)s %(levelname)-8s [Cirro CLI] %(message)s'
)
logger = logging.getLogger("CLI")
logger.setLevel(logging.INFO)
console_handler = logging.StreamHandler()
console_handler.setFormatter(log_formatter)
logger.addHandler(console_handler)


def run_list_datasets(input_params: ListArguments, interactive=False):
    """List the datasets available in a particular project."""
    _check_configure()
    _check_version()
    cirro = CirroApi()
    logger.info(f"Collecting data from {cirro.configuration.base_url}")
    projects = cirro.projects.list()

    if len(projects) == 0:
        raise InputError(NO_PROJECTS)

    if interactive:
        # Prompt the user for the project
        input_params = gather_list_arguments(input_params, projects)
    else:
        input_params['project'] = get_id_from_name(projects, input_params['project'])

    # List the datasets available in that project
    datasets = cirro.datasets.list(input_params['project'])

    sorted_datasets = sorted(datasets, key=lambda d: d.created_at, reverse=True)

    import pandas as pd
    df = pd.DataFrame.from_records([d.to_dict() for d in sorted_datasets])
    df = df[['id', 'name', 'description', 'processId', 'status', 'createdBy', 'createdAt']]
    print(df.to_string())


def run_ingest(input_params: UploadArguments, interactive=False):
    _check_configure()
    _check_version()
    cirro = CirroApi()
    logger.info(f"Collecting data from {cirro.configuration.base_url}")
    processes = cirro.processes.list(process_type=Executor.INGEST)

    logger.info("Listing available projects")
    projects = cirro.projects.list()

    if len(projects) == 0:
        raise InputError(NO_PROJECTS)

    if interactive:
        input_params, files = gather_upload_arguments(input_params, projects, processes)
        directory = input_params['data_directory']
    else:
        input_params['project'] = get_id_from_name(projects, input_params['project'])
        input_params['process'] = get_id_from_name(processes, input_params['process'])
        directory = input_params['data_directory']
        files = get_files_in_directory(directory)

    if len(files) == 0:
        raise InputError("No files to upload")

    process = get_item_from_name_or_id(processes, input_params['process'])
    logger.info(f"Validating expected files: {process.name}")
    try:
        cirro.processes.check_dataset_files(process_id=process.id, files=files, directory=directory)
    except ValueError as e:
        raise InputError(e)
    logger.info("Creating new dataset")

    upload_dataset_request = UploadDatasetRequest(
        process_id=process.id,
        name=input_params['name'],
        description=input_params['description'],
        expected_files=files
    )

    project_id = get_id_from_name(projects, input_params['project'])
    create_resp = cirro.datasets.create(project_id=project_id,
                                        upload_request=upload_dataset_request)

    logger.info("Uploading files")
    cirro.datasets.upload_files(project_id=project_id,
                                dataset_id=create_resp.id,
                                directory=directory,
                                files=files)
    logger.info(f"File content validated by {cirro.configuration.checksum_method_display}")


def run_download(input_params: DownloadArguments, interactive=False):
    _check_configure()
    _check_version()
    cirro = CirroApi()
    logger.info(f"Collecting data from {cirro.configuration.base_url}")

    logger.info("Listing available projects")
    projects = cirro.projects.list()

    if len(projects) == 0:
        raise InputError(NO_PROJECTS)

    files_to_download = None
    if interactive:
        input_params = gather_download_arguments(input_params, projects)

        input_params['project'] = get_id_from_name(projects, input_params['project'])
        datasets = list_all_datasets(project_id=input_params['project'], client=cirro)
        # Filter out datasets that are not complete
        datasets = [d for d in datasets if d.status == Status.COMPLETED]
        input_params = gather_download_arguments_dataset(input_params, datasets)
        files = cirro.datasets.get_assets_listing(input_params['project'], input_params['dataset']).files

        if len(files) == 0:
            raise InputError('There are no files in this dataset to download')

        files_to_download = ask_dataset_files(files)
        project_id = input_params['project']
        dataset_id = input_params['dataset']

    else:
        project_id = get_id_from_name(projects, input_params['project'])
        datasets = cirro.datasets.list(project_id)
        dataset_id = get_id_from_name(datasets, input_params['dataset'])

        if input_params['file']:
            all_files = cirro.datasets.get_assets_listing(project_id, dataset_id).files
            files_to_download = []

            for filepath in input_params['file']:
                if not filepath.startswith('data/'):
                    filepath = os.path.join('data/', filepath)
                file = next((f for f in all_files if f.relative_path == filepath), None)
                if not file:
                    logger.warning(f"Could not find file {filepath}. Skipping.")
                    continue
                files_to_download.append(file)

    logger.info("Downloading files")
    logger.info(f"File content validated by {cirro.configuration.checksum_method_display}")

    cirro.datasets.download_files(project_id=project_id,
                                  dataset_id=dataset_id,
                                  download_location=input_params['data_directory'],
                                  files=files_to_download)


def run_upload_reference(input_params: UploadReferenceArguments, interactive=False):
    _check_configure()
    _check_version()
    cirro = CirroApi()
    logger.info(f"Collecting data from {cirro.configuration.base_url}")

    reference_types = cirro.references.get_types()
    projects = cirro.projects.list()

    if len(projects) == 0:
        raise InputError(NO_PROJECTS)

    if interactive:
        input_params, files = gather_reference_upload_arguments(input_params, projects, reference_types)
    else:
        files = [Path(f) for f in input_params['reference_file']]

    project_id = get_id_from_name(projects, input_params['project'])
    reference_type = next((rt for rt in reference_types if rt.name == input_params['reference_type']), None)

    cirro.references.upload_reference(project_id=project_id,
                                      ref_type=reference_type,
                                      name=input_params['name'],
                                      reference_files=files)


def run_configure():
    _check_version()
    auth_method, base_url, auth_method_config, enable_additional_checksum = gather_auth_config()
    save_user_config(UserConfig(auth_method=auth_method,
                                auth_method_config=auth_method_config,
                                base_url=base_url,
                                transfer_max_retries=None,
                                enable_additional_checksum=enable_additional_checksum))


def run_create_pipeline_config(input_params: CreatePipelineConfigArguments, interactive=False):
    """
    Creates the pipeline configuration files for the CLI.
    This is a placeholder function that can be expanded in the future.
    """
    _check_version()
    logger.info("Creating pipeline configuration files...")

    if interactive:
        input_params = gather_create_pipeline_config_arguments(input_params)
    else:
        if not input_params['pipeline_dir']:
            raise InputError("Root directory is required")
        if not os.path.isdir(input_params['pipeline_dir']):
            raise InputError(f"Root directory {input_params['pipeline_dir']} does not exist")

    logger.debug(input_params)
    pipeline_definition = PipelineDefinition(
        root_dir=input_params['pipeline_dir'],
        entrypoint=input_params.get('entrypoint'),
        logger=logger
    )

    output_dir = input_params.get('output_dir')
    output_paths = {filename: os.path.join(output_dir, filename)  # type: ignore
                    for filename in ['process-form.json', 'process-input.json']}

    logger.info(f"Writing pipeline configuration files to {output_dir}")
    os.makedirs(output_dir, exist_ok=True)

    with open(output_paths['process-form.json'], 'w') as f:
        logger.info(f"Writing form configuration to {output_paths['process-form.json']}")
        json.dump(pipeline_definition.form_configuration, f, indent=2)

    with open(output_paths['process-input.json'], 'w') as f:
        logger.info(f"Writing input configuration to {output_paths['process-input.json']}")
        json.dump(pipeline_definition.input_configuration, f, indent=2)

    logger.info("Pipeline configuration files created successfully.")

    if pipeline_definition.config_app_status == ConfigAppStatus.RECOMMENDED:
        logger.warning(
            "It is recommended that you verify your pipeline configuration "
            "using the Cirro Pipeline Configuration App for this pipeline:\n"
            f"{CONFIG_APP_URL}")


def _check_configure():
    """
    Prompts the user to do initial configuration if needed
    """
    config = load_user_config()
    if config is None:
        run_configure()
        return

    # Legacy check for old config
    if config.base_url == 'cirro.bio':
        run_configure()


def _check_version():
    """
    Prompts the user to update their package version if needed
    """
    yellow_color = '\033[93m'
    reset_color = '\033[0m'

    try:
        current_version = importlib.metadata.version('cirro')
        response = requests.get("https://pypi.org/pypi/cirro/json")
        response.raise_for_status()
        latest_version = response.json()["info"]["version"]

        if current_version != latest_version:
            print(f"{yellow_color}Warning:{reset_color} Cirro version {current_version} "
                  f"is out of date. Update to {latest_version} with 'pip install cirro --upgrade'.")

    except Exception:
        return


def handle_error(e: Exception):
    logger.error(f"{e.__class__.__name__}: {e}")
    sys.exit(1)



from typing import List, Union

from cirro_api_client.v1.models import Process, Executor, ProcessDetail, CustomPipelineSettings, PipelineCode

from cirro.cirro_client import CirroApi
from cirro.models.form_specification import ParameterSpecification
from cirro.sdk.asset import DataPortalAssets, DataPortalAsset


class DataPortalProcess(DataPortalAsset):
    """Helper functions for interacting with analysis processes."""

    def __init__(self, process: Union[Process, ProcessDetail], client: CirroApi):
        """
        Instantiate with helper method

        ```python
        from cirro import DataPortal()
        portal = DataPortal()
        process = portal.get_process_by_name("Process Name")
        ```
        """
        self._data = process
        self._client = client

    @property
    def id(self) -> str:
        """Unique identifier"""
        return self._data.id

    @property
    def name(self) -> str:
        """Readable name"""
        return self._data.name

    @property
    def description(self) -> str:
        """Longer description of process"""
        return self._data.description

    @property
    def child_process_ids(self) -> List[str]:
        """List of processes which can be run on the output of this process"""
        return self._data.child_process_ids

    @property
    def executor(self) -> Executor:
        """INGEST, CROMWELL, or NEXTFLOW"""
        return self._data.executor

    @property
    def category(self) -> str:
        """Category of process"""
        return self._data.category

    @property
    def pipeline_type(self) -> str:
        """Pipeline type"""
        return self._data.pipeline_type

    @property
    def documentation_url(self) -> str:
        """Documentation URL"""
        return self._data.documentation_url

    @property
    def file_requirements_message(self) -> str:
        """Description of files required for INGEST processes"""
        return self._data.file_requirements_message

    @property
    def code(self) -> PipelineCode:
        """Pipeline code configuration"""
        return self._get_detail().pipeline_code

    @property
    def custom_settings(self) -> CustomPipelineSettings:
        """Custom settings for the process"""
        return self._get_detail().custom_settings

    def _get_detail(self) -> ProcessDetail:
        if not isinstance(self._data, ProcessDetail):
            self._data = self._client.processes.get(self.id)
        return self._data

    def __str__(self):
        return '\n'.join([
            f"{i.title()}: {self.__getattribute__(i)}"
            for i in ['name', 'id', 'description']
        ])

    def get_parameter_spec(self) -> ParameterSpecification:
        """
        Gets a specification used to describe the parameters used in the process.
        """
        return self._client.processes.get_parameter_spec(self.id)


class DataPortalProcesses(DataPortalAssets[DataPortalProcess]):
    """Collection of DataPortalProcess objects."""
    asset_name = "process"



from typing import List

from cirro_api_client.v1.api.metrics import get_project_metrics, get_all_metrics
from cirro_api_client.v1.models import ProjectMetrics

from cirro.services.base import BaseService


class MetricsService(BaseService):
    """
    Service for interacting with the Metrics endpoints
    """
    def get_for_project(self, project_id: str) -> ProjectMetrics:
        """
        Retrieves the cost and storage metrics for a project.

        Args:
            project_id (str): ID of the Project
        """
        return get_project_metrics.sync(
            project_id=project_id,
            client=self._api_client
        )

    def get_all_metrics(self) -> List[ProjectMetrics]:
        """
        Retrieves all available metrics
        """
        return get_all_metrics.sync(client=self._api_client)



from cirro_api_client import CirroApiClient

from cirro.auth import get_auth_info_from_config
from cirro.auth.base import AuthInfo
from cirro.config import AppConfig
from cirro.services import FileService, DatasetService, ProjectService, ProcessService, ExecutionService, \
    MetricsService, MetadataService, BillingService, ReferenceService, UserService, ComputeEnvironmentService, \
    ShareService


class CirroApi:
    """
    Client for interacting directly with the Cirro API
    """
    def __init__(self, auth_info: AuthInfo = None, base_url: str = None):
        """
        Instantiates the Cirro API object

        Args:
            auth_info (cirro.auth.base.AuthInfo):
            base_url (str): Optional base URL of the Cirro instance
             (if not provided, it uses the `CIRRO_BASE_URL` environment variable, or the config file)

        Returns:
            Authenticated Cirro API object, which can be used to call endpoint functions.

        Example:
        ```python
        from cirro.cirro_client import CirroApi

        cirro = CirroApi(base_url="app.cirro.bio")
        print(cirro.projects.list())
        ```
        """

        self._configuration = AppConfig(base_url=base_url)
        if not auth_info:
            auth_info = get_auth_info_from_config(self._configuration, auth_io=None)

        self._api_client = CirroApiClient(
            base_url=self._configuration.rest_endpoint,
            auth_method=auth_info.get_auth_method(),
            client_name='Cirro SDK',
            package_name='cirro'
        )

        # Init services
        self._file_service = FileService(self._api_client,
                                         checksum_method=self._configuration.checksum_method,
                                         transfer_retries=self._configuration.transfer_max_retries)
        self._dataset_service = DatasetService(self._api_client, file_service=self._file_service)
        self._project_service = ProjectService(self._api_client)
        self._process_service = ProcessService(self._api_client)
        self._execution_service = ExecutionService(self._api_client)
        self._compute_environment_service = ComputeEnvironmentService(self._api_client)
        self._metrics_service = MetricsService(self._api_client)
        self._metadata_service = MetadataService(self._api_client)
        self._billing_service = BillingService(self._api_client)
        self._references_service = ReferenceService(self._api_client, file_service=self._file_service)
        self._shares_service = ShareService(self._api_client)
        self._users_service = UserService(self._api_client)

    @property
    def datasets(self) -> DatasetService:
        """
        Create, list, delete, and modify Datasets
        """
        return self._dataset_service

    @property
    def projects(self) -> ProjectService:
        """
        Create, list, delete, and modify Projects
        """
        return self._project_service

    @property
    def processes(self) -> ProcessService:
        """
        List and retrieve detailed information about Processes
        """
        return self._process_service

    @property
    def execution(self) -> ExecutionService:
        """
        List, run, stop, and describe the analysis jobs (executing Processes to create new Datasets)
        """
        return self._execution_service

    @property
    def compute_environments(self) -> ComputeEnvironmentService:
        """
        List and update compute environments
        """
        return self._compute_environment_service

    @property
    def metrics(self) -> MetricsService:
        """
        Project-level summary metrics
        """
        return self._metrics_service

    @property
    def metadata(self) -> MetadataService:
        """
        List and modify Sample metadata or metadata schemas
        """
        return self._metadata_service

    @property
    def billing(self) -> BillingService:
        """
        List and update billing accounts
        """
        return self._billing_service

    @property
    def references(self) -> ReferenceService:
        """
        List References and Reference types
        """
        return self._references_service

    @property
    def shares(self) -> ShareService:
        """
        List, create, update, delete, and subscribe to shares
        """
        return self._shares_service

    @property
    def users(self) -> UserService:
        """
        List and update user information
        """
        return self._users_service

    @property
    def file(self) -> FileService:
        """
        Read, download, and create file objects
        """
        return self._file_service

    @property
    def api_client(self) -> CirroApiClient:
        """
        Gets the underlying API client
        """
        return self._api_client

    @property
    def configuration(self) -> AppConfig:
        """
        Gets the configuration of the instance
        """
        return self._configuration
